/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.gitrepository.service;

import com.sun.istack.NotNull;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import ru.ulstu.extractor.branch.model.Branch;
import ru.ulstu.extractor.branch.model.IndexingStatus;
import ru.ulstu.extractor.branch.service.BranchService;
import ru.ulstu.extractor.commit.model.Commit;
import ru.ulstu.extractor.gitrepository.model.GitRepository;
import ru.ulstu.extractor.gitrepository.repository.GitRepositoryRepository;
import ru.ulstu.extractor.ts.creator.db.DBTimeSeriesCreator;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

@Service
public class IndexService {
    private final static Logger LOG = LoggerFactory.getLogger(IndexService.class);
    private final static int COMMITS_PAGE_SIZE = 10;
    private final GitRepositoryService gitRepositoryService;
    private final GitRepositoryRepository gitRepositoryRepository;
    private final BranchService branchService;
    private final List<DBTimeSeriesCreator> timeSeriesCreators;
    private final static ExecutorService EXECUTOR = Executors.newFixedThreadPool(10);

    public IndexService(GitRepositoryService gitRepositoryService,
                        GitRepositoryRepository gitRepositoryRepository,
                        BranchService branchService,
                        List<DBTimeSeriesCreator> timeSeriesCreators) {
        this.gitRepositoryService = gitRepositoryService;
        this.gitRepositoryRepository = gitRepositoryRepository;
        this.branchService = branchService;
        this.timeSeriesCreators = timeSeriesCreators;
    }

    @Transactional
    @Async
    public void index(@NotNull Integer branchId) throws GitAPIException, IOException {
        Branch branch = branchService.findByBranchId(branchId).orElseThrow(() -> new RuntimeException("Ветка репозитория не найдена по id"));
        index(branch.getGitRepository(), branch);
    }

    @Transactional
    @Async
    public void index(String gitRepositoryUrl, String branchName) throws GitAPIException, IOException {
        GitRepository gitRepository = gitRepositoryService.findByUrlOrCreate(gitRepositoryUrl);
        Branch branch = branchService.findByRepositoryAndNameOrCreate(gitRepository, branchName);
        branch = branchService.updateStatus(branch, IndexingStatus.INDEXING);
        index(gitRepository, branch);
    }

    @Transactional
    @Async
    public void index(GitRepository gitRepository, Branch branch) throws GitAPIException, IOException {
        branchService.save(branch, Collections.emptyList());
        branch = branchService.updateStatus(branch, IndexingStatus.INDEXING);
        int commitsFrom = 0;
        int commitsTo = COMMITS_PAGE_SIZE;
        List<Commit> commits = gitRepositoryService.getCommits(gitRepository.getUrl(), branch.getName(), commitsFrom, commitsTo, true);
        while (!commits.isEmpty()) {
            LOG.debug("{} commits loaded.", commits.size());
            branchService.addCommits(branch, commits);
            LOG.debug("{} commits successfully saved. {} {}", commits.size(), gitRepository.getUrl(), branch.getName());
            commitsFrom += COMMITS_PAGE_SIZE;
            commitsTo += COMMITS_PAGE_SIZE;
            commits = gitRepositoryService.getCommits(gitRepository.getUrl(), branch.getName(), commitsFrom, commitsTo, false);
        }
        final Branch branchForSave = branch;
        timeSeriesCreators.forEach(tsCreator -> tsCreator.addTimeSeries(branchForSave));
        branchService.updateStatus(branch, IndexingStatus.FINISHED);
        LOG.debug("Complete indexing {} branch", branch.getName());
    }

    @Transactional
    @Async
    public void indexFailedBranchesOnStart() {
        LOG.info("Старт проверки незавершенных задач для индексирований...");

        List<Branch> failedBranches = branchService.findAllByIndexingStatus(IndexingStatus.INDEXING);
        if (failedBranches.size() > 0) {
            LOG.info("Найдено {} незавершенных задач для индексирования", failedBranches.size());
            failedBranches.forEach(failedBranch -> {
                try {
                    index(failedBranch.getId());
                    LOG.info("Завершено индексирование ветки {}", failedBranch.getName());
                } catch (Exception ex) {
                    LOG.warn(ex.getMessage());
                }
            });
            LOG.info("Завершено индексирование незавершенных задач");
        } else {
            LOG.info("Не найдено незавершенных веток для индексирования");
        }
    }
}
